using System;
using System.Collections;
using System.Collections.Generic;
using UnityEngine;

public class PlayerController : MonoBehaviour
{
    /* Objects */
    [Header("Objects")] public Transform Head;  /* The location of our "head" (used to see if we can keep moving up */
    
    /* Movemement */
    [Header("Movement")] 
    public float BaseMoveSpeed = 2f;        /* The movement speed of the character */
    public bool IsGrounded;                 /* True if the character is "grounded" and touching the ground */
    public float JumpVelocity = 5f;         /* The amount of velocity a jump imparts vertically */
    public float MaxFallVelocity = 45f;     /* The maximum amount of fall speed a character can have */
    public int MaxJumps = 3;                /* How many jumps can the player do before touching the ground? */

    private CharacterController characterController;    /* The character controller to apply motion to */
    private Vector3 Velocity = Vector3.zero;        /* The velocity of the player this frame */
    private Vector3 baseVelocity = Vector3.zero;    /* The velocity of the platform beneath the player */
    private GameObject currentFloor;                /* The current "floor" we're standing on */
    private int currentJumps = 0;                   /* How many jumps have we done since we were grounded */
    private float groundedJumpTime = 0;             /* The time when the jumper left the ground */
    private float groundedGracePeriod = 0.5f;       /* The period of time after a grounded jump to stop re-grounding */

    private void Awake()
    {
        characterController = GetComponent<CharacterController>();
    }

    // Update is called once per frame
    void Update()
    {
        /* Handle WASD Movement */
        HandleHorizontalMovement();
        
        /* Update the "IsGrounded" variable */
        CheckIfGrounded();
        
        /* Handle Jumping */
        HandleJumping();
        
        /* Handle Vertical Movement */
        HandleVerticalMovement();
        
        /* Handle Platforms */
        HandlePlatformBehavior();
        
        /* Add any other controls/abilities you want! */
        /* Could add a spin move, a dash, or anything else you can think of */
        
        /* Apply the movement */
        characterController.Move((Velocity + baseVelocity) * Time.deltaTime);
        
        /* Rotate the model to face "forward" if we moved */
        Vector3 horizontalVelocity = new Vector3(Velocity.x, 0, Velocity.z).normalized;
        if (horizontalVelocity.sqrMagnitude != 0)
        {
            transform.rotation = Quaternion.LookRotation(horizontalVelocity, Vector3.up);
        }
        
        /* Reset Velocity (but attempt to maintain vertical velocity) */
        float vertical = Mathf.Max(Velocity.y, -MaxFallVelocity);
        Velocity = Vector3.zero;
        Velocity.y = vertical;
    }

    /// <summary>
    /// Handle wasd movement.
    /// </summary>
    void HandleHorizontalMovement()
    {
        /* Receive Vector Movement */
        float horizontal = Input.GetAxisRaw("Horizontal");
        float vertical = Input.GetAxisRaw("Vertical");
        Vector2 moveVec = new Vector2(horizontal, vertical).normalized;
        
        /* Get camera direction and "flatten" it */
        Vector3 cameraForward = Camera.main.transform.forward;
        cameraForward.y = 0;
        cameraForward.Normalize();

        /* Move the character with respect to the camera's direction */
        /* Forward = away from camera */
        Vector3 forwardDir = moveVec.y * cameraForward;
        Velocity += forwardDir * BaseMoveSpeed;

        /* Horizontal = side from camera */
        Vector3 rightDir = -Vector3.Cross(cameraForward, Vector3.up);
        Velocity += rightDir * BaseMoveSpeed * moveVec.x;
    }

    /// <summary>
    /// Handle operations related to vertical movement, like
    /// gravity and detecting hitting your head.
    /// </summary>
    void HandleVerticalMovement()
    {
        /* Apply Gravity */
        Velocity += Physics.gravity * Time.deltaTime;   // we need another delta time here: remember, gravity is acceleration!
        
        /* Check if we're moving upwards and hitting our head - if we are, we need to stop trying to move upwards */
        if (Velocity.y > 0)
        {
            if (Physics.OverlapBox(Head.position, Vector3.one * 0.5f).Length > 2)
            {
                /* we're moving upwards and hitting our head: zero that velocity! */
                Velocity.y = 0;
            }
        }
    }

    /// <summary>
    /// Updates whether or not the character is grounded.+
    /// </summary>
    void CheckIfGrounded()
    {
        /* First, don't consider grounding right after a jump (to avoid getting "stuck" */
        if (Time.time - groundedJumpTime < groundedGracePeriod) return;
        
        /* Raycast downwards to see if a collider is hit */
        if (Physics.Raycast(transform.position, Vector3.down, out var hitinfo, 1.1f))
        {
            /* We have a hit and must be grounded. Update vars */
            IsGrounded = true;
            currentFloor = hitinfo.transform.gameObject;
            
            /* If we're grounded, our vertical velocity should be zero */
            Velocity.y = 0;
            
            /* We can also reset our jump counter */
            currentJumps = 0;
        }
        else
        {
            /* No hit, we must be airborne */
            IsGrounded = false;
            currentFloor = null;
        }
    }

    /// <summary>
    /// Handle any jumping behavior.
    /// </summary>
    void HandleJumping()
    {
        if (Input.GetKeyDown(KeyCode.Space))
        {
            if (currentJumps < MaxJumps)
            {
                Velocity.y = JumpVelocity;  /* Jumping should reset velocity completely */
                currentJumps += 1;
                if (IsGrounded)
                {
                    groundedJumpTime = Time.time;
                }
            }
        }
    }

    /// <summary>
    /// Apply platform velocity to the character.
    /// DOES NOT WORK WITH ROTATION!
    /// </summary>
    void HandlePlatformBehavior()
    {
        if (currentFloor != null)
        {
            if (currentFloor.TryGetComponent(out VelocityReader vr))
            {
                baseVelocity = vr.GetVelocity();
            }
            else
            {
                baseVelocity = Vector3.zero;
            }
        }
        else
        {
            baseVelocity = Vector3.zero;
        }
    }
}
